/*! \file
   This header defines utility functions that help to avoid code duplication among 
   different techniques for integrating single scattering in participating media.*/

/*! Converts the given rectified coordinates to Cartesian coordinates in 
   rectification space and returns the result.
  \sa CartesianToRectified() */
float3 RectifiedToCartesian(float Distance,float Inclination,float Azimuth){
	float AzimuthSine,AzimuthCosine;
	sincos(Azimuth,AzimuthSine,AzimuthCosine);
	float InclinationSine,InclinationCosine;
	sincos(Inclination,InclinationSine,InclinationCosine);
	return Distance*float3(AzimuthCosine,AzimuthSine,InclinationCosine/InclinationSine);
}


/*! This function converts the given position in Cartesian coordinates of 
   rectification space into rectified coordinates. It is a right inverse of 
   RectifiedToCartesian().
  \param OutDistance The non-negative distance from the origin after projecting to 
		 the z=0 plane.
  \param OutInclination The angle to the z-axis in radians from zero to pi.
  \param OutAzimuth The planar angle to the x-axis in radians from zero to two pi.
  \param CartesianCoordinates The Cartesian coordinates that are to be converted.*/
void CartesianToRectified(out float OutDistance,out float OutInclination,out float OutAzimuth,float3 CartesianCoordinates){
	const float PI=3.1415926535897932384626433832795f;
	OutDistance=length(CartesianCoordinates.xy);
	OutInclination=atan2(OutDistance,CartesianCoordinates.z);
	OutAzimuth=atan2(CartesianCoordinates.y,CartesianCoordinates.x);
	OutAzimuth=(OutAzimuth<0.0f)?(OutAzimuth+2.0f*PI):OutAzimuth;
}


/*! This function constructs a linear combination of two Dirac-delta distributions 
   which matches three given moments.
  \param OutDepth The locations of the Dirac-delta distributions. They are sorted.
  \param OutWeight1 The weight of the second Dirac-delta. The weights add up to one.
  \param OptimizedMoments Four moments stored in a manner that is optimized for 
		 quantization. The fourth moment is not used.*/
void GetSparseRepresentation4Moments(out float2 OutDepth,out float OutWeight1,float4 OptimizedMoments){
	// Get the first three moments
	OptimizedMoments[0]-=0.035955884801f;
	float3 b=mul(OptimizedMoments,float4x3(
			0.2227744146f, 0.1549679261f, 0.1451988946f,
			0.0771972861f, 0.1394629426f, 0.2120202157f,
			0.7926986636f, 0.7963415838f, 0.7258694464f,
			0.0319417555f,-0.1722823173f,-0.2758014811f));
	// Determine the locations of the Dirac-delta pulses
	float3 Polynomial;
	float Variance=mad(-b.x,b.x,b.y);
	Polynomial.z=Variance;
	Polynomial.y=mad(b.x,b.y,-b.z);
	Polynomial.x=dot(-b.xy,Polynomial.yz);
	float Scaling=1.0f/Polynomial[2];
	float p=Polynomial[1]*Scaling;
	float q=Polynomial[0]*Scaling;
	float D=p*p*0.25f-q;
	float r=sqrt(D);
	OutDepth=float2(-0.5f*p-r,-0.5f*p+r);
	// Get the weight
	OutWeight1=(b.x-OutDepth[0])/(OutDepth[1]-OutDepth[0]);
	// Negative variance may occur due to quantization errors. This is easily and 
	// appropriately handled using a single Dirac-delta at the mean
	bool SingleDiracDelta=(Variance<=1.0e-6f);
	OutDepth=SingleDiracDelta?float2(b.x,0.0f):OutDepth;
	OutWeight1=SingleDiracDelta?0.0f:OutWeight1;
}


/*! Estimates the integral of a probability distribution from minus infinity to a 
   fixed point and returns it.
  \param IntervalEnd The end of the integration interval.
  \param OptimizedMoments Power moments one to four of the distribution stored in a 
		 a manner that is optimized for little quantization errors.
  \param MomentBias The strength of the bias that serves to compensate quantization 
		 errors (e.g. 3*10^-5).
  \param OverestimationWeight If this value is zero, the function underestimates the 
		 integral. If it is one, it overestimates. Inbetween linear interpolation is 
		 employed.*/
float EstimateIntegralFrom4Moments(float IntervalEnd,float4 OptimizedMoments,float MomentBias,float OverestimationWeight){
	OptimizedMoments[0]-=0.035955884801f;
	float4 _4Moments=mul(OptimizedMoments,float4x4(
			0.2227744146f, 0.1549679261f, 0.1451988946f, 0.163127443f,
			0.0771972861f, 0.1394629426f, 0.2120202157f, 0.2591432266f,
			0.7926986636f, 0.7963415838f, 0.7258694464f, 0.6539092497f,
			0.0319417555f,-0.1722823173f,-0.2758014811f,-0.3376131734f));

	// Bias input data to avoid artifacts
	const float4 BiasVector=lerp(float4(0.5f,1.0f/3.0f,0.25f,0.2f),float4(0.5f,0.5f,0.5f,0.5f),0.8f);
	float4 b=lerp(_4Moments,BiasVector,MomentBias);
	float3 z;
	z[0]=IntervalEnd;

	// Compute a Cholesky factorization of the Hankel matrix B storing only non-
	// trivial entries or related products
	float L21D11=mad(-b[0],b[1],b[2]);
	float D11=mad(-b[0],b[0], b[1]);
	float SquaredDepthVariance=mad(-b[1],b[1], b[3]);
	float D22D11=dot(float2(SquaredDepthVariance,-L21D11),float2(D11,L21D11));
	float InvD11=1.0f/D11;
	float L21=L21D11*InvD11;

	// Obtain a scaled inverse image of bz=(1,z[0],z[0]*z[0])^T
	float3 c=float3(1.0f,z[0],z[0]*z[0]);
	// Forward substitution to solve L*c1=bz
	c[1]-=b.x;
	c[2]-=b.y+L21*c[1];
	// Scaling to solve D*c2=c1
	c[1]*=InvD11;
	c[2]*=D11/D22D11;
	// Backward substitution to solve L^T*c3=c2
	c[1]-=L21*c[2];
	c[0]-=dot(c.yz,b.xy);
	// Solve the quadratic equation c[0]+c[1]*z+c[2]*z^2 to obtain solutions 
	// z[1] and z[2]
	float p=c[1]/c[2];
	float q=c[0]/c[2];
	float D=(p*p*0.25f)-q;
	float r=sqrt(D);
	z[1]=-p*0.5f-r;
	z[2]=-p*0.5f+r;
	// Compute the shadow intensity by summing the appropriate weights
	float3 Weight;
	Weight[0]=(z[1]*z[2]-b[0]*(z[1]+z[2])+b[1])/((z[0]-z[1])*(z[0]-z[2]));
	Weight[1]=(z[0]*z[2]-b[0]*(z[0]+z[2])+b[1])/((z[2]-z[1])*(z[0]-z[1]));
	Weight[2]=1.0f-Weight[0]-Weight[1];
	float Integral=
		(z[2]<z[0])?(Weight[1]+Weight[2]):(
		(z[1]<z[0])?Weight[1]:0.0f);
	Integral+=Weight[0]*OverestimationWeight;
	return saturate(Integral);
}

/*! This utility function computes one cubic root of the given complex value. The 
   other roots can be found by multiplication by cubic roots of unity.
  \note This function has various discontinuities.*/
float2 ComplexCubicRoot(float2 Z){
	float Argument=atan2(Z.y,Z.x);
	float NewArgument=Argument/3.0f;
	float2 NormalizedRoot;
	sincos(NewArgument,NormalizedRoot.y,NormalizedRoot.x);
	return NormalizedRoot*pow(dot(Z,Z),1.0f/6.0f);
}

/*! Given the coefficients of a real, cubic polynomial Coefficient[0]
  +x*Coefficient[1]+x*x*Coefficient[2]+x*x*x*Coefficient[3] this function computes 
  all of its roots and returns them. Roots with multiplicity are returned multiple 
  times. The roots are returned in increasing order.
 \warning Behavior is undefined if the polynomial does not have three real roots.*/
float3 GetRoots(float4 Coefficient){
	// This implementation is based upon the algorithm and notions on Wikipedia:
	// http://en.wikipedia.org/wiki/Cubic_polynomial#General_formula_for_roots
	float bb=Coefficient[2]*Coefficient[2];
	float ac=Coefficient[1]*Coefficient[3];
	float Delta0=mad(-3.0f,ac,bb);
	float bbb=bb*Coefficient[2];
	float abc=ac*Coefficient[2];
	float aad=Coefficient[3]*Coefficient[3]*Coefficient[0];
	float Delta1=dot(float3(2.0f,-9.0f,27.0f),float3(bbb,abc,aad));
	float aInv=1.0f/Coefficient[3];
	float2 CCube=0.5f*float2(Delta1,sqrt(4.0f*Delta0*Delta0*Delta0-Delta1*Delta1));
	float2 C=ComplexCubicRoot(CCube);
	float2 Delta0OverC=Delta0*float2(C.x,-C.y)/dot(C,C);
	float3 Root;
	Root[0]=Coefficient[2]+C.x+Delta0OverC.x;
	float RootMiddle=Coefficient[2]-0.5f*(C.x+Delta0OverC.x);
	Root[1]=RootMiddle-0.5f*sqrt(3.0f)*(C.y-Delta0OverC.y);
	Root[2]=RootMiddle+0.5f*sqrt(3.0f)*(C.y-Delta0OverC.y);
	Root*=-(1.0f/3.0f)*aInv;
	// Do a few steps of Newton's method as post-iteration
	[unroll] for(int i=0;i!=1;++i){
		float3 Polynomial=mad(mad(mad(Coefficient[3],Root,Coefficient[2]),Root,Coefficient[1]),Root,Coefficient[0]);
		float3 Derivative=mad(mad(3.0f*Coefficient[3],Root,2.0f*Coefficient[2]),Root,Coefficient[1]);
		Root-=Polynomial/Derivative;
	}
	return Root;
}

/*! Like EstimateIntegralFrom4Moments() but uses six moments which have been stored 
   in a manner that is optimized for small quantization error.*/
float EstimateIntegralFrom6Moments(float IntervalEnd,float2 OptimizedMoments0,float4 OptimizedMoments1,float MomentBias,float OverestimationWeight){
	// Dequantize the moments
	OptimizedMoments0-=float2(9.9994469329e-01f,3.6984078658e-04f);
	OptimizedMoments1.yzw-=float3(8.0785948090e-05f,5.9277028569e-06f,9.7746026694e-01f);
	const float2x2 QuantizationMatrixLeftTop=float2x2(
				-0.7923732636f,-0.7927589967f,
				 0.2341398592f, 0.2338879746f);
	const float2x4 QuantizationMatrixRightTop=float2x4(
				-0.7339425714f,-0.6748893652f,-0.6236866843f,-0.580409736f,
				 0.2799720667f, 0.3271482957f, 0.3652961959f, 0.3940151284f);
	const float4x2 QuantizationMatrixLeftBottom=float4x2(
				 0.0364202291f,-0.0057927764f,
				-0.0263111927f,-0.0268873079f,
				 0.005106059f,  0.0038746589f,
				-0.0064527671f, 0.2100390127f);
	const float4x4 QuantizationMatrixRightBottom=float4x4(
				-0.0232180143f,-0.0301246087f,-0.032463882f, -0.0336729124f,
				-0.0143681921f,-0.0025883286f, 0.0104261903f, 0.024972809f,
				 0.0085756361f,-0.0019812868f,-0.0213251514f,-0.0436908639f,
				 0.3231516668f, 0.3820444425f, 0.4129392296f, 0.4285574001f);
	float2 b12=mul(OptimizedMoments0,QuantizationMatrixLeftTop)+mul(OptimizedMoments1,QuantizationMatrixLeftBottom);
	float4 b3456=mul(OptimizedMoments0,QuantizationMatrixRightTop)+mul(OptimizedMoments1,QuantizationMatrixRightBottom);
	float b[6]={b12.x,b12.y,b3456.x,b3456.y,b3456.z,b3456.w};
	// Bias input data to avoid artifacts
	[unroll] for(int i=0;i!=6;++i){
		const float BiasVectorEntry=lerp(1.0f/float(i+2),0.5f,0.5f);
		b[i]=lerp(b[i],BiasVectorEntry,MomentBias);
	}
	float4 z;
	z[0]=IntervalEnd;
	
	// Compute a Cholesky factorization of the Hankel matrix B storing only non-
	// trivial entries or related products
    float InvD11=1.0f/mad(-b[0],b[0],b[1]);
    float L21D11=mad(-b[0],b[1],b[2]);
    float L21=L21D11*InvD11;
	float D22=mad(-L21D11,L21,mad(-b[1],b[1],b[3]));
    float L31D11=mad(-b[0],b[2],b[3]);
    float L31=L31D11*InvD11;
    float InvD22=1.0f/D22;
    float L32D22=mad(-L21D11,L31,mad(-b[1],b[2],b[4]));
    float L32=L32D22*InvD22;
	float D33=mad(-b[2],b[2],b[5])-dot(float2(L31D11,L32D22),float2(L31,L32));
    float InvD33=1.0f/D33;

	// Construct the polynomial whose roots have to be points of support of the 
	// canonical distribution: bz=(1,z[0],z[0]*z[0],z[0]*z[0]*z[0])^T
	float4 c;
	c[0]=1.0f;
	c[1]=z[0];
	c[2]=c[1]*z[0];
	c[3]=c[2]*z[0];
	// Forward substitution to solve L*c1=bz
	c[1]-=b[0];
	c[2]-=mad(L21,c[1],b[1]);
	c[3]-=b[2]+dot(float2(L31,L32),c.yz);
	// Scaling to solve D*c2=c1
	c.yzw*=float3(InvD11,InvD22,InvD33);
	// Backward substitution to solve L^T*c3=c2
	c[2]-=L32*c[3];
	c[1]-=dot(float2(L21,L31),c.zw);
	c[0]-=dot(float3(b[0],b[1],b[2]),c.yzw);

	// Solve the cubic equation c[0]+c[1]*z+c[2]*z^2+c[3]*z^3 to obtain solutions 
	// z[1], z[2] and z[3]
	z.yzw=GetRoots(c);
	// Determine how the weights at these roots contribute to the end result
	float4 WeightFactor;
	WeightFactor[0]=1.0f-OverestimationWeight;
	WeightFactor.yzw=(z.yzw>z.xxx)?float3(1.0f,1.0f,1.0f):float3(0.0f,0.0f,0.0f);
	// Construct an interpolation polynomial matching these factors at the Roots
	float f0=WeightFactor[0];
	float f1=WeightFactor[1];
	float f2=WeightFactor[2];
	float f3=WeightFactor[3];
	float f01=(f1-f0)/(z[1]-z[0]);
	float f12=(f2-f1)/(z[2]-z[1]);
	float f23=(f3-f2)/(z[3]-z[2]);
	float f012=(f12-f01)/(z[2]-z[0]);
	float f123=(f23-f12)/(z[3]-z[1]);
	float f0123=(f123-f012)/(z[3]-z[0]);
	// Convert from the Newton basis to the canonical basis of polynomials
	float4 InterpolationPolynomial;
	// f012+f0123*(z-z2)
	InterpolationPolynomial[0]=mad(-f0123,z[2],f012);
	InterpolationPolynomial[1]=f0123;
	// *(z-z1)  +f01
	InterpolationPolynomial[2]=InterpolationPolynomial[1];
	InterpolationPolynomial[1]=mad(InterpolationPolynomial[1],-z[1],InterpolationPolynomial[0]);
	InterpolationPolynomial[0]=mad(InterpolationPolynomial[0],-z[1],f01);
	// *(z-z0)  +f0
	InterpolationPolynomial[3]=InterpolationPolynomial[2];
	InterpolationPolynomial[2]=mad(InterpolationPolynomial[2],-z[0],InterpolationPolynomial[1]);
	InterpolationPolynomial[1]=mad(InterpolationPolynomial[1],-z[0],InterpolationPolynomial[0]);
	InterpolationPolynomial[0]=mad(InterpolationPolynomial[0],-z[0],f0);

	// Now the end result happens to be the dot product of the first four moments 
	// and the interpolation polynomial
	return 1.0f-saturate(dot(InterpolationPolynomial,float4(1.0f,b[0],b[1],b[2])));
}


/*! This is a helper function for functions generating transmittance weighted shadow 
   maps with resampling of a shadow map for non-linear rectification. It precomputes 
   various useful quantities that can be prepared before entering a loop over a row 
   of a shadow map.
  \param OutTextureSpaceRayStart,OutTextureSpaceRayOffset Coordinates defining the 
		 ray which is to be traced in the source shadow map. They use projective 
		 coordinates, you should divide by z to get the texture coordinate in xy. 
		 The offset is the offset from one output shadow map column to the next.
  \param OutShadowMapToRectificationDepthTransform A linear transform mapping shadow 
		 map depth to depth in rectification space. x is the factor, y the summand.
  \param OutInclinationToDepthTransform A linear transform mapping inclination in 
		 radians to a value that can be used as depth value for the rectified shadow 
		 map.
  \param OutOffsetLength The length of the component of the offset defined by 
		 OutTextureSpaceRayOffset which is orthogonal to the light direction. The 
		 length is measured with the canonical metric of world space.
  \param OutInitialOffsetLength The minimal distance between the light ray identified 
		 by OutTextureSpaceRayStart and the camera position in world space.
  \param OutputShadowMapResolution The resolution of the shadow map in which prefix 
		 sums will be stored.
  \param iShadowMapRow The index of the row in the output shadow map (i.e. the 
		 non-negative integer y-coordinate).
  \param RectificationToShadowMapProjectionSpace See 
		 ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4().
  \param DistanceMinMax,InclinationMinMax,AzimuthMinMax See 
		 GetRectifiedSpaceFrustumBounds().
  \sa ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4() */
void PrepareShadowMapRowResampling(out float3 OutTextureSpaceRayStart,out float3 OutTextureSpaceRayOffset,out float2 OutShadowMapToRectificationDepthTransform,out float2 OutInclinationToDepthTransform,out float OutOffsetLength,out float OutInitialOffsetLength,uint2 OutputShadowMapResolution,uint iShadowMapRow,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	// Construct the reference ray in the epipolar slice
	float RayAzimuth=lerp(AzimuthMinMax.x,AzimuthMinMax.y,float(iShadowMapRow)/float(OutputShadowMapResolution.y));
	float ReferenceInclination=dot(InclinationMinMax,float2(0.5f,0.5f));
	float4 RectificationSpaceRayDirection=float4(RectifiedToCartesian(1.0f,ReferenceInclination,RayAzimuth),0.0f);
	float4 ProjectionSpaceRayDirection=mul(RectificationSpaceRayDirection,RectificationToShadowMapProjectionSpace);
	float4 ProjectionSpaceRayStart=mul(float4(0.0f,0.0f,0.0f,1.0f),RectificationToShadowMapProjectionSpace);
	// Convert to a coordinate system where shadow map texture coordinates arise 
	// directly to save one multiply-add within the loop
	float3 TextureSpaceRayDirection=ProjectionSpaceRayDirection.xyw*float3(0.5f,-0.5f,1.0f)+ProjectionSpaceRayDirection.www*float3(0.5f,0.5f,0.0f);
	// Prepare coefficients to map shadow map depth to rectification space depth
	float4 ProjectionSpaceCameraPosition=mul(float4(0.0f,0.0f,0.0f,1.0f),RectificationToShadowMapProjectionSpace);
	float CameraShadowMapDepth=ProjectionSpaceCameraPosition.z/ProjectionSpaceCameraPosition.w;
	float4 ProjectionSpaceNorthPole=mul(float4(0.0f,0.0f,1.0f,1.0f),RectificationToShadowMapProjectionSpace);
	float NorthPoleShadowMapDepth=ProjectionSpaceNorthPole.z/ProjectionSpaceNorthPole.w;
	OutShadowMapToRectificationDepthTransform.x=1.0f/(NorthPoleShadowMapDepth-CameraShadowMapDepth);
	OutShadowMapToRectificationDepthTransform.y=-CameraShadowMapDepth*OutShadowMapToRectificationDepthTransform.x;
	// Also prepare coefficients for conversion of inclinations to depth
	OutInclinationToDepthTransform.x=1.0f/(InclinationMinMax.y-InclinationMinMax.x);
	OutInclinationToDepthTransform.y=-InclinationMinMax.x*OutInclinationToDepthTransform.x;
	// Compute the offset length and the final offset
	OutOffsetLength=DistanceMinMax.y/float(OutputShadowMapResolution.x);
	OutInitialOffsetLength=-0.5f*OutOffsetLength;
	OutTextureSpaceRayOffset=TextureSpaceRayDirection*OutOffsetLength;
	OutTextureSpaceRayStart=ProjectionSpaceRayStart.xyw*float3(0.5f,-0.5f,1.0f)+ProjectionSpaceRayStart.www*float3(0.5f,0.5f,0.0f);
	OutTextureSpaceRayStart-=0.5f*OutTextureSpaceRayOffset;
}


/*! This function converts a depth take from a common shadow map and maps it to a 
   depth value as used for non-linear rectification.
  \param ShadowMapDepth The depth value obtained from the shadow map.
  \param ShadowMapToRectificationDepthTransform,InclinationToDepthTransform See 
		 PrepareShadowMapRowResampling().
  \param TotalOffsetLength The number of times the offset has been applied to the 
		 ray start times the length of the offset as generated by 
		 PrepareShadowMapRowResampling().
  \return The resampled depth value. It is obtained from a shadow map sample at the 
		  given texture coordinate using a non-linear mapping.*/
float ConvertShadowMapDepthNonLinearRectification(float ShadowMapDepth,float2 ShadowMapToRectificationDepthTransform,float2 InclinationToDepthTransform,float TotalOffsetLength){
	float RectificationSpaceDepth=mad(ShadowMapDepth,ShadowMapToRectificationDepthTransform.x,ShadowMapToRectificationDepthTransform.y);
	float SampleInclination=atan2(TotalOffsetLength,RectificationSpaceDepth);
	return saturate(mad(SampleInclination,InclinationToDepthTransform.x,InclinationToDepthTransform.y));
}


/*! Converts a given depth value to a vector encoding six moments of a Dirac-delta 
   distribution at this depth. The moments are output as two vectors and an affine 
   transform is applied to them to prepare them for storage in a texture with 
   minimal loss of information due to quantization.*/
void ComputeMomentVector6MomentsOptimized(out float2 OutQuantizedMoments0,out float4 OutQuantizedMoments1,float Depth){
	// Generate the moments
	float SquaredDepth=Depth*Depth;
	float CubedDepth=Depth*SquaredDepth;
	float2 Moments0=float2(Depth,SquaredDepth);
	float4 Moments1=float4(CubedDepth,SquaredDepth*SquaredDepth,CubedDepth*SquaredDepth,CubedDepth*CubedDepth);
	// Apply the affine transform
	const float2x2 QuantizationMatrixLeftTop=float2x2(
		 -3.9982174417e-01f,-2.0364580370e-01f,
		  5.3724994330e+00f, 6.1687438614e+01f);
	const float2x4 QuantizationMatrixRightTop=float2x4(
		 3.5997693025e+01f, 2.4253667437e+01f, 1.4571662352e+01f, 2.5464326857e+00f,
		-4.1996056141e+02f,-1.8746243964e+02f,-6.4120119623e+01f,-7.8049820130e+01f);
	const float4x2 QuantizationMatrixLeftBottom=float4x2(
		-1.9400931206e+01f,-2.8026108456e+02f,
		 5.9601506283e+00f, 4.8086777452e+02f,
		 2.1716342727e+01f,-3.7061702860e+02f,
		-1.4248134400e+01f, 1.0952605942e+02f);
	const float4x4 QuantizationMatrixRightBottom=float4x4(
		 1.7918207666e+03f, 4.9910228252e+02f, 7.5292307128e+01f, 3.5077691236e+02f,
		-3.4556578547e+03f,-5.1652881873e+02f, 8.7981549920e+00f,-6.7456153641e+02f,
		 3.0717060681e+03f, 1.4463179523e+02f,-4.3073239823e+01f, 5.9868902736e+02f,
		-1.0239061080e+03f, 3.7003250859e+01f, 8.5315616209e+00f,-1.9939897152e+02f);
	OutQuantizedMoments0=mul(Moments0,QuantizationMatrixLeftTop)+mul(Moments1,QuantizationMatrixLeftBottom);
	OutQuantizedMoments1=mul(Moments0,QuantizationMatrixRightTop)+mul(Moments1,QuantizationMatrixRightBottom);
	OutQuantizedMoments0+=float2(9.9994469329e-01f,3.6984078658e-04f);
	OutQuantizedMoments1.yzw+=float3(8.0785948090e-05f,5.9277028569e-06f,9.7746026694e-01f);
}


/*! Converts a given depth value to a vector encoding four moments of a Dirac-delta 
   distribution at this depth. An affine transform is applied to the moments to 
   prepare them for storage in a texture with minimal loss of information due to 
   quantization.*/
void ComputeMomentVector4MomentsOptimized(out float4 OutQuantizedMoments,float Depth){
	float SquaredDepth=Depth*Depth;
	OutQuantizedMoments=float4(Depth,SquaredDepth,SquaredDepth*Depth,SquaredDepth*SquaredDepth);
	OutQuantizedMoments=mul(OutQuantizedMoments,float4x4(
					-2.07224649f,   13.7948857237f,  0.105877704f,   9.7924062118f,
					32.23703778f,  -59.4683975703f, -1.9077466311f,-33.7652110555f,
				   -68.571074599f,  82.0359750338f,  9.3496555107f, 47.9456096605f,
					39.3703274134f,-35.364903257f,  -6.6543490743f,-23.9728048165f));
	OutQuantizedMoments[0]+=0.035955884801f;
}
