/*! Given a quality factor (with default value 1.0) this function computes quality 
   coefficients. These are used by ApplyParallaxConeStep(...).*/
void ProcessParallaxQuality(out float4 OutParallaxQualityCoefficients,float ParallaxQuality){
	float InvParallaxQuality=1.0f/ParallaxQuality;
	OutParallaxQualityCoefficients.x=0.02f*InvParallaxQuality;
	OutParallaxQualityCoefficients.y=0.002f*InvParallaxQuality;
	OutParallaxQualityCoefficients.z=0.2f;
	OutParallaxQualityCoefficients.w=1.0f;
}


/*! Given distances at which the decrease of the parallax intensity starts and ends, 
   this function computes corresponding coefficients, which can be used by 
   ApplyParallaxConeStep(...).*/
void ProcessParallaxDistance(out float2 OutParallaxDistanceCoefficients,float ParallaxDecreaseStartDistance,float ParallaxDecreaseEndDistance){
	float DecreaseStartDistanceSq=ParallaxDecreaseStartDistance*ParallaxDecreaseStartDistance;
	float DecreaseEndDistanceSq=ParallaxDecreaseEndDistance*ParallaxDecreaseEndDistance;
	float Factor=1.0f/(DecreaseStartDistanceSq-DecreaseEndDistanceSq);
	OutParallaxDistanceCoefficients.x=Factor;
	OutParallaxDistanceCoefficients.y=-Factor*DecreaseEndDistanceSq;
}


/*! This function computes the mesh to tangent space transformation and outputs it 
   as 3x3 matrix.*/
void ComputeMeshToTangentSpace(out float3x3 OutMeshToTangentSpace,float4 TangentSpaceData,float3 MeshSpaceNormal){
	// Compute the mesh to tangent space
	float3 MeshSpaceTangent=TangentSpaceData.xyz;
	float3 MeshSpaceBitangent=cross(MeshSpaceNormal,MeshSpaceTangent)/TangentSpaceData.w;
	OutMeshToTangentSpace=float3x3(
		MeshSpaceTangent.x,MeshSpaceBitangent.x,MeshSpaceNormal.x,
		MeshSpaceTangent.y,MeshSpaceBitangent.y,MeshSpaceNormal.y,
		MeshSpaceTangent.z,MeshSpaceBitangent.z,MeshSpaceNormal.z);
}


/*! This function prepares the data, which is needed for simple parallax mapping and 
   passes it to the pixel shader.*/
void PrepareParallaxSimple(out float2 OutParallaxDataSimple,float3x3 MeshToTangentSpace,float3 WorldSpaceCameraPosition,float3 WorldSpacePosition,float MaterialDepth,float4x4 WorldToMeshSpace){
	float3 MSViewVector=mul(WorldSpacePosition-WorldSpaceCameraPosition,(float3x3)WorldToMeshSpace);
	float3 TSViewVector=mul(MSViewVector,MeshToTangentSpace);
	TSViewVector.z=min(-MaterialDepth*5.0f,TSViewVector.z);
	TSViewVector*=-MaterialDepth/TSViewVector.z;
	TSViewVector*=0.1f*min(10.0f,rsqrt(dot(TSViewVector,TSViewVector)));
	OutParallaxDataSimple=TSViewVector.xy;
}

/*! This function takes a texture coordinate and a parallax offset and uses both to 
   perform simple parallax mapping using the given height map.*/
void ApplyParallaxSimple(out float2 OutTexCoord,float2 OriginalTexCoord,float2 ParallaxDataSimple,SSamplerTexturePair2D HeightMapSampler){
	float Height=HeightMapSampler.Texture.Sample(HeightMapSampler.Sampler,OriginalTexCoord).a;
	OutTexCoord=OriginalTexCoord+ParallaxDataSimple*(1.0f-Height);
}


/*! This function prepares the data, which is needed for regula falsi parallax 
   mapping and passes it to the pixel shader.*/
void PrepareParallaxRF(out float2 OutParallaxDataRF,float2 OriginalTexCoord,float3x3 MeshToTangentSpace,float3 WorldSpaceCameraPosition,float3 WorldSpacePosition,float MaterialDepth,float4x4 WorldToMeshSpace){
	float3 MSViewVector=mul(WorldSpacePosition-WorldSpaceCameraPosition,(float3x3)WorldToMeshSpace);
	float3 TSViewVector=mul(MSViewVector,MeshToTangentSpace);
	TSViewVector.z=min(-MaterialDepth*5.0f,TSViewVector.z);
	TSViewVector*=-MaterialDepth/TSViewVector.z;
	TSViewVector*=0.1f*min(10.0f,rsqrt(dot(TSViewVector,TSViewVector)));
	OutParallaxDataRF=OriginalTexCoord.xy+TSViewVector.xy;
}

/*! This function takes a texture coordinate and an offset texture coordinate and uses 
   both to perform regula falsi parallax mapping.*/
void ApplyParallaxRF(out float2 OutTexCoord,float2 OriginalTexCoord,float2 ParallaxDataRF,SSamplerTexturePair2D HeightMapSampler){
	float Height0=HeightMapSampler.Texture.Sample(HeightMapSampler.Sampler,OriginalTexCoord).a;
	float Height1=HeightMapSampler.Texture.Sample(HeightMapSampler.Sampler,ParallaxDataRF).a;
	float NegYIntercept=1.0f-Height0;
	float Zero=NegYIntercept/(Height1+NegYIntercept);
	OutTexCoord=lerp(OriginalTexCoord,ParallaxDataRF,Zero.xx);
}


/*! This function prepares the data, which is needed to do cone step parallax mapping 
   and passes it to the pixel shader.*/
void PrepareParallaxConeStep(out float4 OutParallaxDataConeStep,float3x3 MeshToTangentSpace,float3 WorldSpaceCameraPosition,float4x4 WorldToMeshSpace,float3 WorldSpacePosition,float MaterialDepth){
	// Compute the tangent space view vector
	float3 WSViewVector=WorldSpacePosition.xyz-WorldSpaceCameraPosition;
	float3 MSViewVector=mul(WSViewVector,(float3x3)WorldToMeshSpace);
	float4 TSViewVector;
	TSViewVector.xyz=mul(MSViewVector,MeshToTangentSpace);
	// An alternative z-coordinate is needed. The normal points inside and has the 
	// length MaterialDepth.
	TSViewVector.w=max(MaterialDepth,-TSViewVector.z)/MaterialDepth;
	// Pass the result to the pixel shader. Avoid that a mesh scaling effects 
	// distances.
	OutParallaxDataConeStep=TSViewVector;
	// Compute the squared distance from the camera to the vertex. It is used for 
	// disabling parallax mapping at the appropriate distance.
	OutParallaxDataConeStep.z=dot(WSViewVector,WSViewVector);
}

/*! This function takes a texture coordinate and an offset texture coordinate and uses 
   both to perform cone step parallax mapping.*/
void ApplyParallaxConeStep(out float2 OutTexCoord,float2 OriginalTexCoord,float4 ParallaxDataConeStep,SSamplerTexturePair2D HeightMapSampler,float MaterialDepth,float4 ParallaxQualityCoefficients,float2 ParallaxDistanceCoefficients){
	// The parallax intensity gets reduced at higher distances
	float DistanceSq=ParallaxDataConeStep.z;
	ParallaxDataConeStep.xy*=saturate(DistanceSq*ParallaxDistanceCoefficients.x+ParallaxDistanceCoefficients.y);
	// The maximal angle of the view ray gets limited
	ParallaxDataConeStep.w=max(ParallaxDataConeStep.w,sqrt(ParallaxQualityCoefficients.w*dot(ParallaxDataConeStep.xy,ParallaxDataConeStep.xy)));
	// Compute the intersection of the view ray with the lower end of the simulated 
	// volume
	ParallaxDataConeStep.xy/=ParallaxDataConeStep.w;
	// Compute the partial derivative of the input texture coordinate to have a 
	// reasonable gradient for height map sampling
	float2 DerivativeUVX=ddx(OriginalTexCoord);
	float2 DerivativeUVY=ddy(OriginalTexCoord);
	// Compute a minimal step width to avoid too long iterations
	float InvStepCount=lerp(ParallaxQualityCoefficients.x,ParallaxQualityCoefficients.y,rsqrt(dot(ParallaxDataConeStep.xyz,ParallaxDataConeStep.xyz)));
	// Start the iterative search for an intersection between the height map and the 
	// view ray
	float4 HeightMapData=HeightMapSampler.Texture.Sample(HeightMapSampler.Sampler,OriginalTexCoord);
	float ConeRadiusFactor=sqrt(max(0.00001f,dot(ParallaxDataConeStep.xy,ParallaxDataConeStep.xy)));
	float IntervalEnd=0.0f;
	float Height0=0.0f;
	float ViewRayY=1.0f;
	float DeltaX=0.0f;
	// The iteration ends when the view ray has intersected the height map
	while(ViewRayY>HeightMapData.w){
		// Compute the next texture coordinate using the cone step data
		Height0=HeightMapData.w;
		float ConeFactor=HeightMapData.z/(ConeRadiusFactor+HeightMapData.z);
		DeltaX=min(ParallaxQualityCoefficients.z,max(InvStepCount,(ViewRayY-HeightMapData.w)*ConeFactor));
		IntervalEnd+=DeltaX;
		// Sample at this position
		HeightMapData=tex2Dgrad(HeightMapSampler,OriginalTexCoord+ParallaxDataConeStep.xy*IntervalEnd,DerivativeUVX,DerivativeUVY);
		// Compute the new height value of the view ray
		ViewRayY=1.0f-IntervalEnd;
	}
	// To improve the result at the end a regula falsi step is added
	float DeltaY=HeightMapData.w-Height0;
	float Root=(DeltaX-Height0*DeltaX+DeltaY*(IntervalEnd-DeltaX))/(DeltaY+DeltaX);
	// Now the final texture coordinate can be computed
	OutTexCoord=OriginalTexCoord+ParallaxDataConeStep.xy*Root;
}
