/*! \file
   This file defines functions to define and render render participating media. Most 
   notably absorption and single scattering can be rendered.*/

#include "RandomNumbers.fx"
#include "ParticipatingMediaUtility.fx"

/*! This function prepares a matrix that is useful when working with non-linear 
   rectification transforms for directional lights and prefiltered single 
   scattering.
  \param OutRectificationToWorldSpace The rectification to world space transform. 
		 Rectification space is essentially light view space with the world space 
		 camera position at its origin. Additionally, the z and y axis are mirrored 
		 to ensure that inclinations grow as the depth grows. From rectification 
		 space coordinates rectified coordinates can be computed immediately.
  \param WorldSpaceCameraPosition The position of the camera for which single 
		 scattering is to be rendered in world space.
  \param ShadowMapWorldToViewSpace The world to view space transform used for the 
		 camera that renders the shadow map of the relevant directional light. It is 
		 assumed to be a rotation.*/
void ComputeRectificationToWorldSpaceDirectional(out float4x4 OutRectificationToWorldSpace,float3 WorldSpaceCameraPosition,float4x4 ShadowMapViewToWorldSpace){
	OutRectificationToWorldSpace=float4x4(
		 ShadowMapViewToWorldSpace._11, ShadowMapViewToWorldSpace._12, ShadowMapViewToWorldSpace._13, 0.0f,
		-ShadowMapViewToWorldSpace._21,-ShadowMapViewToWorldSpace._22,-ShadowMapViewToWorldSpace._23, 0.0f,
		-ShadowMapViewToWorldSpace._31,-ShadowMapViewToWorldSpace._32,-ShadowMapViewToWorldSpace._33, 0.0f,
		 WorldSpaceCameraPosition.x,    WorldSpaceCameraPosition.y,    WorldSpaceCameraPosition.z,    1.0f
	);
}


/*! This function computes the coordinate ranges covered by the rectified 
   coordinates of the specified view frustum.
  \param OutDistanceMinMax x is always set to zero, y is set to the maximal distance 
		 between the camera position and a point within the frustum.
  \param OutInclinationMinMax Minimal and maximal inclination of points within the 
		 frustum in rectification space. This ranges from 0 to pi.
  \param OutAzimuthMinMax Minimal and maximal azimuth of points within the frustum 
		 in rectification space. x and y-x are both always between 0 and 2*pi. The 
		 interval of radian angles inbetween covers the frustum.
  \param CameraProjectionToWorldSpace The projection to world space transform of the 
		 camera for which single scattering is to be rendered. This defines the 
		 frustum.
  \param CameraWorldToProjectionSpace Inverse of CameraProjectionToWorldSpace.
  \param RectificationToWorldSpace See 
		 ComputeRectificationToWorldSpaceDirectional().
  \param WorldToRectificationSpace Inverse of RectificationToWorldSpace.*/
void GetRectifiedSpaceFrustumBounds(out float2 OutDistanceMinMax,out float2 OutInclinationMinMax,out float2 OutAzimuthMinMax,float4x4 CameraProjectionToWorldSpace,float4x4 CameraWorldToProjectionSpace,float4x4 RectificationToWorldSpace,float4x4 WorldToRectificationSpace){
	const float PI=3.1415926535897932384626433832795f;
	// Compute the rectification space corners of the far plane
	float4 pFarPlaneCorner[4]={
		float4(-1.0f,-1.0f,1.0f,1.0f),
		float4(-1.0f, 1.0f,1.0f,1.0f),
		float4( 1.0f, 1.0f,1.0f,1.0f),
		float4( 1.0f,-1.0f,1.0f,1.0f)
	};
	[unroll] for(int i=0;i!=4;++i){
		pFarPlaneCorner[i]=mul(pFarPlaneCorner[i],CameraProjectionToWorldSpace);
		pFarPlaneCorner[i]=mul(pFarPlaneCorner[i],WorldToRectificationSpace);
		pFarPlaneCorner[i]/=pFarPlaneCorner[i].w;
	}
	// Compute the maximal distance
	OutDistanceMinMax=float2(0.0f,0.0f);
	[unroll] for(int i=0;i!=4;++i){
		OutDistanceMinMax.y=max(OutDistanceMinMax.y,length(pFarPlaneCorner[i].xy));
	}
	// Figure out whether one of the poles of the unit sphere lies between the side 
	// clipping planes
	float4 NorthPole=mul(mul(float4(0.0f,0.0f,1.0f,1.0f),RectificationToWorldSpace),CameraWorldToProjectionSpace);
	bool4 ClipResult=bool4(NorthPole.x<=NorthPole.w,NorthPole.x>=-NorthPole.w,NorthPole.y<=NorthPole.w,NorthPole.y>=-NorthPole.w);
	bool NorthPoleInFrustum=all(ClipResult);
	bool SouthPoleInFrustum=all(!ClipResult);
	// Compute the azimuths
	if(NorthPoleInFrustum || SouthPoleInFrustum){
		OutAzimuthMinMax=float2(0.0f,2.0f*PI);
	}
	else{
		// Project down to the plane and normalize
		float2 pFarPlaneCornerFlat[4];
		[unroll] for(int i=0;i!=4;++i){
			pFarPlaneCornerFlat[i]=normalize(pFarPlaneCorner[i].xy);
		}
		// Compute the biggest pairwise angle
		float MinDotProduct=2.0f;
		float2 pCorner[2];
		[unroll] for(int i=0;i!=4;++i){
			[unroll] for(int j=0;j!=i;++j){
				float DotProduct=dot(pFarPlaneCornerFlat[i],pFarPlaneCornerFlat[j]);
				bool IsMinimal=(DotProduct<MinDotProduct);
				pCorner[0]=IsMinimal?pFarPlaneCornerFlat[i]:pCorner[0];
				pCorner[1]=IsMinimal?pFarPlaneCornerFlat[j]:pCorner[1];
				MinDotProduct=IsMinimal?DotProduct:MinDotProduct;
			}
		}
		// Figure out which of the two corners is clockwise of the other (when going 
		// no farther than pi)
		bool FirstIsClockwise=dot(float2(-pCorner[0].y,pCorner[0].x),pCorner[1])>0.0f;
		// Compute the angles
		float AzimuthRange=acos(MinDotProduct);
		float2 MinCorner=FirstIsClockwise?pCorner[0]:pCorner[1];
		OutAzimuthMinMax.x=atan2(MinCorner.y,MinCorner.x);
		OutAzimuthMinMax.x=(OutAzimuthMinMax.x<0.0f)?(OutAzimuthMinMax.x+2.0f*PI):OutAzimuthMinMax.x;
		OutAzimuthMinMax.y=OutAzimuthMinMax.x+AzimuthRange;
	}
	// Compute the inclinations. Start with the corner points.
	OutInclinationMinMax=float2(PI,0.0f);
	[unroll] for(int i=0;i!=4;++i){
		float CornerInclination=acos(normalize(pFarPlaneCorner[i].xyz).z);
		OutInclinationMinMax.x=min(OutInclinationMinMax.x,CornerInclination);
		OutInclinationMinMax.y=max(OutInclinationMinMax.y,CornerInclination);
	}
	// Now handle the case that one of the lines bounding the far plane holds an 
	// extremum
	[unroll] for(int i=0;i!=4;++i){
		float3 Start=pFarPlaneCorner[i].xyz;
		float3 End=pFarPlaneCorner[(i<3)?(i+1):0].xyz;
		float3 Offset=End-Start;
		// Determine the location of the extremal inclination on the line
		float InclinationExtremum=(Start.z*dot(Start,Offset)-Offset.z*dot(Start,Start))
								 /(Offset.z*dot(Start,Offset)-Start.z*dot(Offset,Offset));
		InclinationExtremum=saturate(InclinationExtremum);
		// Compute the inclination and update the minima and maxima
		float3 ExtremumLocation=normalize(lerp(Start,End,InclinationExtremum));
		float ExtremalLineInclination=acos(ExtremumLocation.z);
		OutInclinationMinMax.x=min(OutInclinationMinMax.x,ExtremalLineInclination);
		OutInclinationMinMax.y=max(OutInclinationMinMax.y,ExtremalLineInclination);
	}
	// Finally, handle the case that some point in the inner of the far plane is an 
	// extremum
	OutInclinationMinMax.x=NorthPoleInFrustum?0.0f:OutInclinationMinMax.x;
	OutInclinationMinMax.y=SouthPoleInFrustum?PI:OutInclinationMinMax.y;
	// Stretch inclination bounds a little to avoid unnecessary leaking
	OutInclinationMinMax=lerp(OutInclinationMinMax.xx,OutInclinationMinMax.yy,float2(-0.05f,1.05f));
	OutInclinationMinMax=clamp(OutInclinationMinMax,0.0f,PI);
}


/*! @{
   This function loads the samples of the given texture for the specified texel and 
   takes their average. This is a somewhat reasonable custom resolve for a depth 
   stencil texture in the context of single scattering because single scattering 
   depends monotonously (although not linearly) on depth.*/
void GetPerspectiveDepth1(out float OutPerspectiveDepth,Texture2DMS<float,1> DepthStencilTexture,int2 TexelIndex){
	uint Width,Height,nSample;
	DepthStencilTexture.GetDimensions(Width,Height,nSample);
	OutPerspectiveDepth=0.0f;
	[unroll] for(uint i=0;i!=nSample;++i){
		OutPerspectiveDepth+=DepthStencilTexture.Load(TexelIndex,i);
	}
	OutPerspectiveDepth/=nSample;
}
void GetPerspectiveDepth2(out float OutPerspectiveDepth,Texture2DMS<float,2> DepthStencilTexture,int2 TexelIndex){
	uint Width,Height,nSample;
	DepthStencilTexture.GetDimensions(Width,Height,nSample);
	OutPerspectiveDepth=0.0f;
	[unroll] for(uint i=0;i!=nSample;++i){
		OutPerspectiveDepth+=DepthStencilTexture.Load(TexelIndex,i);
	}
	OutPerspectiveDepth/=nSample;
}
void GetPerspectiveDepth4(out float OutPerspectiveDepth,Texture2DMS<float,4> DepthStencilTexture,int2 TexelIndex){
	uint Width,Height,nSample;
	DepthStencilTexture.GetDimensions(Width,Height,nSample);
	OutPerspectiveDepth=0.0f;
	[unroll] for(uint i=0;i!=nSample;++i){
		OutPerspectiveDepth+=DepthStencilTexture.Load(TexelIndex,i);
	}
	OutPerspectiveDepth/=nSample;
}
void GetPerspectiveDepth8(out float OutPerspectiveDepth,Texture2DMS<float,8> DepthStencilTexture,int2 TexelIndex){
	uint Width,Height,nSample;
	DepthStencilTexture.GetDimensions(Width,Height,nSample);
	OutPerspectiveDepth=0.0f;
	[unroll] for(uint i=0;i!=nSample;++i){
		OutPerspectiveDepth+=DepthStencilTexture.Load(TexelIndex,i);
	}
	OutPerspectiveDepth/=nSample;
}
//! @}


/*! This function computes the extinction coefficient of participating media from 
   the given half opacity distance.*/
void ComputeExtinctionCoefficient(out float OutExtinctionCoefficient,float HalfOpacityDistance){
	OutExtinctionCoefficient=-log(0.5f)/HalfOpacityDistance;
}


/*! This function computes the world space distance between the given camera 
   position and fragment position.*/
void ComputeViewRayLengthFromWorldSpace(out float OutViewRayLength,float3 WorldSpacePosition,float3 WorldSpaceCameraPosition){
	OutViewRayLength=distance(WorldSpacePosition,WorldSpaceCameraPosition);
}


/*! This function converts an outgoing radiance to an incoming radiance by taking 
   the attenuation due to absorption in homogenous participating media into account.
  \param OutReceivedRadiance The radiance at the end of the considered ray.
  \param OutgoingRadiance The radiance at the beginning of the ray.
  \param ViewRayLength The length of the considered ray.
  \param ExtinctionCoefficient The extinction coefficient.*/
void ApplyAbsorptionToOutgoingRadiance(out float3 OutReceivedRadiance,float3 OutgoingRadiance,float ViewRayLength,float ExtinctionCoefficient){
	OutReceivedRadiance=OutgoingRadiance*exp(-ExtinctionCoefficient*ViewRayLength);
}


/*! This function computes the distance between a specified point in projection 
   space and the origin of the view space coordinate system (which is supposed to be 
   the camera location) in view space.
  \param OutViewRayLength The computed view space distance.
  \param OutWorldSpaceRayStart The world space position at which the ray starts 
		 (i.e. the camera position).
  \param OutWorldSpaceRayEnd The world space position at which the ray ends (i.e. 
		 the specified point).
  \param TexCoord The texture coordinate that corresponds to the point of interest 
		 in the render target.
  \param PerspectiveDepth The depth stored in the depth stencil texture for the 
		 point of interest. It is perspectivic (z/w) not linear (z).
  \param ProjectionToViewSpace The projection to view space transform of the 
		 camera.
  \param ViewToWorldSpace The view to world space transform of the camera.*/
void ComputeViewRayFromProjectionSpace(out float OutViewRayLength,out float3 OutWorldSpaceRayStart,out float3 OutWorldSpaceRayEnd,float2 TexCoord,float PerspectiveDepth,float4x4 ProjectionToViewSpace,float4x4 ViewToWorldSpace){
	float4 ProjectionSpacePosition=float4(TexCoord*float2(2.0f,-2.0f)+float2(-1.0f,1.0f),PerspectiveDepth,1.0f);
	float4 ViewSpacePosition=mul(ProjectionSpacePosition,ProjectionToViewSpace);
	ViewSpacePosition/=ViewSpacePosition.w;
	OutWorldSpaceRayStart=mul(float4(0.0f,0.0f,0.0f,1.0f),ViewToWorldSpace).xyz;
	OutWorldSpaceRayEnd=mul(ViewSpacePosition,ViewToWorldSpace).xyz;
	OutViewRayLength=length(ViewSpacePosition.xyz);
}


/*! This function computes the constant factors in front of the single scattering 
   integral divided by the extinction coefficient. This is simply the product of its 
   arguments.*/
void ComputeSingleScatteringFactors(out float3 OutPhaseTimesAlbedoTimesIrradiance,float3 PhaseTimesAlbedo,float3 DirectionalLightIrradiance){
	OutPhaseTimesAlbedoTimesIrradiance=PhaseTimesAlbedo*DirectionalLightIrradiance;
}

/*! Given the value of the single scattering integral times the extinction 
   coefficient this function computes the radiance received due to single 
   scattering. The value of the scattering integral has to be a value between zero 
   (no scattering) and one (maximal scattering).
  \sa ComputeSingleScatteringFactors() */
void ComputeSingleScatteringRadiance(out float3 OutScatteringRadiance,float ScatteringIntegral,float3 PhaseTimesAlbedoTimesIrradiance){
	OutScatteringRadiance=ScatteringIntegral*PhaseTimesAlbedoTimesIrradiance;
}

/*! Given the radiance value due to single scattering this function computes the 
   color to be output by the shader for additive blending. Essentially, it just sets 
   the alpha to one.
  \sa ComputeSingleScatteringRadiance() */
void ComputeScatteringColor(out float4 OutScatteringColor,float3 ScatteringRadiance){
	OutScatteringColor.rgb=ScatteringRadiance;
	OutScatteringColor.a=1.0f;
}


/*! This function evaluates the single scattering integral under the assumption that 
   the light is never occluded.
  \param OutScatteringIntegral See ComputeSingleScatteringRadiance().
  \param ViewRayLength The length of the view ray for which single scattering should 
		 be accumulated.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.*/
void ComputeSingleScatteringNoOcclusionDirectional(out float OutScatteringIntegral,float ViewRayLength,float ExtinctionCoefficient){
	OutScatteringIntegral=1.0f-exp(-ExtinctionCoefficient*ViewRayLength);
}

/*! This function evaluates the scattering integral resolving occlusions by sampling 
   the given shadow map densely with some jittering to replace banding by noise.
  \param OutScatteringIntegral See ComputeSingleScatteringRadiance().
  \param ViewRayLength The length of the view ray for which single scattering should 
		 be accumulated.
  \param WorldSpaceViewRayStart,WorldSpaceViewRayEnd The endpoints of the view ray 
		 in world space coordinates.
  \param ShadowMapSampler A sampler for a shadow map of the scene.
  \param ShadowMapWorldToProjectionSpace The world to projection space transform of 
		 the provided shadow map.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.*/
void ComputeSingleScatteringRayMarchingDirectional(out float OutScatteringIntegral,float ViewRayLength,float3 WorldSpaceViewRayStart,float3 WorldSpaceViewRayEnd,SSamplerTexturePair2D_float ShadowMapSampler,float4x4 ShadowMapWorldToProjectionSpace,float ExtinctionCoefficient){
	OutScatteringIntegral=0.0f;
	const unsigned int nSample=128;
	// Convert ray start and ray end to projection space and compute the offset 
	// between consecutive samples
	float4 RayStart=mul(float4(WorldSpaceViewRayStart,1.0f),ShadowMapWorldToProjectionSpace);
	float4 RayEnd=mul(float4(WorldSpaceViewRayEnd,1.0f),ShadowMapWorldToProjectionSpace);
	// Convert the projection space coordinates to coordinates that enable immediate 
	// computation of texture coordinates to save two mads within the loop
	RayStart.xy=RayStart.xy*float2(0.5f,-0.5f)+RayStart.w*float2(0.5f,0.5f);
	RayEnd.xy=RayEnd.xy*float2(0.5f,-0.5f)+RayEnd.w*float2(0.5f,0.5f);
	// Implement jittering to avoid banding. Do this using a cheap linear 
	// congruential generator.
	float InvSampleCount=1.0f/nSample;
	float RandomOffset=GetRandomNumber(WorldSpaceViewRayEnd);
	float InitialLerpFactor=-RandomOffset*InvSampleCount;
	// Prepare variables that can be updated efficiently during iterations of the 
	// loop
	float4 SampleOffset=(RayEnd-RayStart)*InvSampleCount;
	float4 SamplePosition=lerp(RayStart,RayEnd,InitialLerpFactor);
	float Weight=exp(-ExtinctionCoefficient*InitialLerpFactor*ViewRayLength)-exp(-ExtinctionCoefficient*(InitialLerpFactor+InvSampleCount)*ViewRayLength);
	float WeightFactor=exp(-ExtinctionCoefficient*ViewRayLength*InvSampleCount);
	for(int i=0;i!=nSample;++i){
		// Compute the sample position in shadow map projection space
		SamplePosition+=SampleOffset;
		// Do the look up in the shadow map
		float InvSamplePositionW=1.0f/SamplePosition.w;
		float2 TexCoord=SamplePosition.xy*InvSamplePositionW;
		float RayDepth=SamplePosition.z*InvSamplePositionW;
		float ShadowMapDepth=ShadowMapSampler.Texture.SampleLevel(ShadowMapSampler.Sampler,TexCoord,0.0f);
		// Compute the contribution to the single scattering integral
		bool LightVisible=(RayDepth<ShadowMapDepth || TexCoord.x<0.0f || TexCoord.y<0.0f || TexCoord.x>1.0f || TexCoord.y>1.0f);
		Weight*=WeightFactor;
		OutScatteringIntegral+=LightVisible?Weight:0.0f;
	}
}


/*! This function computes coordinates which are needed for evaluation of 
   prefiltered single scattering with any prefiltered, linearly rectified shadow 
   map.
  \param OutRayEndTexCoord The texture coordinate at which the prefiltered shadow 
		 map has to be sampled to get information about the provided ray.
  \param OutRayDepth The depth of the relevant ray in shadow map coordinates.
  \param WorldSpaceViewRayEnd The endpoint of the view ray in world space 
		 coordinates.
  \param ShadowMapWorldToProjectionSpace The world to projection space transform of 
		 the relevant shadow map.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.
  \sa ApplyTransmittanceWeightedPrefixSumLinearRectification() */
void ComputePrefilteredSingleScatteringCoordinatesLinearRectification(out float2 OutRayEndTexCoord,out float OutRayDepth,float3 WorldSpaceViewRayEnd,float4x4 ShadowMapWorldToProjectionSpace){
	// Determine where to sample the moment shadow map
	float4 ProjectionSpaceRayEnd=mul(float4(WorldSpaceViewRayEnd,1.0f),ShadowMapWorldToProjectionSpace);
	ProjectionSpaceRayEnd/=ProjectionSpaceRayEnd.w;
	OutRayEndTexCoord=ProjectionSpaceRayEnd.xy*float2(0.5f,-0.5f)+float2(0.5f,0.5f);
	OutRayDepth=ProjectionSpaceRayEnd.z;
}


/*! Like ComputePrefilteredSingleScatteringCoordinatesLinearRectification() but for 
   non-linearly rectified shadow maps.
  \param OutRayEndTexCoord,WorldSpaceViewRayEnd See 
		 ComputePrefilteredSingleScatteringCoordinatesLinearRectification().
  \param OutRayDepth The inclination of the relevant ray. It functions as shadow map 
		 depth value in the rectified coordinate system.
  \param WorldToRectificationSpace Inverse of the transform provided by 
		 ComputeRectificationToWorldSpaceDirectional().
  \param DistanceMinMax,InclinationMinMax,AzimuthMinMax See 
		 GetRectifiedSpaceFrustumBounds().
  \sa ApplyTransmittanceWeightedPrefixSumNonLinearRectification() */
void ComputePrefilteredSingleScatteringCoordinatesNonLinearRectification(out float2 OutRayEndTexCoord,out float OutRayDepth,float3 WorldSpaceViewRayEnd,float4x4 WorldToRectificationSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	const float PI=3.1415926535897932384626433832795f;
	// Compute rectified coordinates for the view ray end in rectification space
	float4 RectificationSpaceViewRayEnd=mul(float4(WorldSpaceViewRayEnd,1.0f),WorldToRectificationSpace);
	float Distance,Inclination,Azimuth;
	CartesianToRectified(Distance,Inclination,Azimuth,RectificationSpaceViewRayEnd.xyz);
	// Normalize the coordinates
	float InvMaxDistance=1.0f/DistanceMinMax.y;
	OutRayEndTexCoord.x=Distance*InvMaxDistance;
	OutRayEndTexCoord.y=Azimuth-AzimuthMinMax.x;
	OutRayEndTexCoord.y=(OutRayEndTexCoord.y<0.0f)?(OutRayEndTexCoord.y+2.0f*PI):OutRayEndTexCoord.y;
	float InvAzimuthRange=1.0f/(AzimuthMinMax.y-AzimuthMinMax.x);
	OutRayEndTexCoord.y*=InvAzimuthRange;
	OutRayDepth=(Inclination-InclinationMinMax.x)/(InclinationMinMax.y-InclinationMinMax.x);
}


/*! This function evaluates the scattering integral resolving occlusions by means of 
   a rectified moment shadow map holding weighted prefix sums of rows.
  \param OutScatteringIntegral See ComputeSingleScatteringRadiance().
  \param RayEndTexCoord,RayDepth See 
		 ComputePrefilteredSingleScatteringCoordinatesLinearRectification().
  \param ScatteringIntegralWithoutOcclusion The output value in absence of shadow.
  \param ShadowMapSampler0,ShadowMapSampler1,ShadowMapSampler2,ShadowMapSampler3 
		 Samplers for the rectified and prefiltered shadow maps holding  of the 
		 scene.*/
void ComputePrefilteredSingleScatteringConvolution8(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,SSamplerTexturePair2D ShadowMapSampler2,SSamplerTexturePair2D ShadowMapSampler3){
	const float PI=3.1415926535897932384626433832795f;
	// Get the Fourier coefficients from the textures
	float4 Fourier1_2=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	float4 Fourier3_4=ShadowMapSampler1.Texture.Sample(ShadowMapSampler1.Sampler,RayEndTexCoord);
	float4 Fourier5_6=ShadowMapSampler2.Texture.Sample(ShadowMapSampler2.Sampler,RayEndTexCoord);
	float4 Fourier7_8=ShadowMapSampler3.Texture.Sample(ShadowMapSampler3.Sampler,RayEndTexCoord);
	float2 pFourierCoefficient[8]={
		Fourier1_2.xy*2.0f-1.0f,
		Fourier1_2.zw*2.0f-1.0f,
		Fourier3_4.zw*2.0f-1.0f,
		Fourier3_4.zw*2.0f-1.0f,
		Fourier5_6.zw*2.0f-1.0f,
		Fourier5_6.zw*2.0f-1.0f,
		Fourier7_8.zw*2.0f-1.0f,
		Fourier7_8.zw*2.0f-1.0f
	};
	// Evaluate the first complex basis function
	float2 Fourier1;
	sincos(RayDepth*PI,Fourier1.y,Fourier1.x);
	// Evaluate the other basis functions using complex exponentiation
	float2 pBasisFunction[8];
	pBasisFunction[0]=Fourier1;
	float2 FourierSquare=float2(dot(Fourier1,float2(Fourier1.x,-Fourier1.y)),dot(Fourier1.xy,Fourier1.yx));
	[unroll] for(int i=1;i!=8;++i){
		float2 LHS=pBasisFunction[i-1];
		float2 RHS=FourierSquare;
		pBasisFunction[i]=float2(dot(LHS,float2(RHS.x,-RHS.y)),dot(LHS.xy,RHS.yx));
	}
	// Compute the shadow intensity which is a linear combination of the basis 
	// functions using the Fourier coefficients as weights
	OutScatteringIntegral=0.5f;
	[unroll] for(int i=0;i!=8;++i){
		float Constant=PI*(2.0f*i+1.0f);
		float Summand=pBasisFunction[i].y*pFourierCoefficient[i].x-pBasisFunction[i].x*pFourierCoefficient[i].y;
		OutScatteringIntegral+=Summand*2.0f/Constant;
	}
	OutScatteringIntegral=saturate(1.0f-OutScatteringIntegral);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! Like ComputePrefilteredSingleScatteringConvolution16() but for eight Fourier 
   coefficients.*/
void ComputePrefilteredSingleScatteringConvolution16(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,SSamplerTexturePair2D ShadowMapSampler2,SSamplerTexturePair2D ShadowMapSampler3,SSamplerTexturePair2D ShadowMapSampler4,SSamplerTexturePair2D ShadowMapSampler5,SSamplerTexturePair2D ShadowMapSampler6,SSamplerTexturePair2D ShadowMapSampler7){
	const float PI=3.1415926535897932384626433832795f;
	// Get the Fourier coefficients from the textures
	float4 Fourier1_2=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	float4 Fourier3_4=ShadowMapSampler1.Texture.Sample(ShadowMapSampler1.Sampler,RayEndTexCoord);
	float4 Fourier5_6=ShadowMapSampler2.Texture.Sample(ShadowMapSampler2.Sampler,RayEndTexCoord);
	float4 Fourier7_8=ShadowMapSampler3.Texture.Sample(ShadowMapSampler3.Sampler,RayEndTexCoord);
	float4 Fourier9_10=ShadowMapSampler4.Texture.Sample(ShadowMapSampler4.Sampler,RayEndTexCoord);
	float4 Fourier11_12=ShadowMapSampler5.Texture.Sample(ShadowMapSampler5.Sampler,RayEndTexCoord);
	float4 Fourier13_14=ShadowMapSampler6.Texture.Sample(ShadowMapSampler6.Sampler,RayEndTexCoord);
	float4 Fourier15_16=ShadowMapSampler7.Texture.Sample(ShadowMapSampler7.Sampler,RayEndTexCoord);
	float2 pFourierCoefficient[16]={
		Fourier1_2.xy*2.0f-1.0f,
		Fourier1_2.zw*2.0f-1.0f,
		Fourier3_4.zw*2.0f-1.0f,
		Fourier3_4.zw*2.0f-1.0f,
		Fourier5_6.zw*2.0f-1.0f,
		Fourier5_6.zw*2.0f-1.0f,
		Fourier7_8.zw*2.0f-1.0f,
		Fourier7_8.zw*2.0f-1.0f,
		Fourier9_10.xy*2.0f-1.0f,
		Fourier9_10.zw*2.0f-1.0f,
		Fourier11_12.xy*2.0f-1.0f,
		Fourier11_12.zw*2.0f-1.0f,
		Fourier13_14.xy*2.0f-1.0f,
		Fourier13_14.zw*2.0f-1.0f,
		Fourier15_16.xy*2.0f-1.0f,
		Fourier15_16.zw*2.0f-1.0f
	};
	// Evaluate the first complex basis function
	float2 Fourier1;
	sincos(RayDepth*PI,Fourier1.y,Fourier1.x);
	// Evaluate the other basis functions using complex exponentiation
	float2 pBasisFunction[16];
	pBasisFunction[0]=Fourier1;
	float2 FourierSquare=float2(dot(Fourier1,float2(Fourier1.x,-Fourier1.y)),dot(Fourier1.xy,Fourier1.yx));
	[unroll] for(int i=1;i!=16;++i){
		float2 LHS=pBasisFunction[i-1];
		float2 RHS=FourierSquare;
		pBasisFunction[i]=float2(dot(LHS,float2(RHS.x,-RHS.y)),dot(LHS.xy,RHS.yx));
	}
	// Compute the shadow intensity which is a linear combination of the basis 
	// functions using the Fourier coefficients as weights
	OutScatteringIntegral=0.5f;
	[unroll] for(int i=0;i!=16;++i){
		float Constant=PI*(2.0f*i+1.0f);
		float Summand=pBasisFunction[i].y*pFourierCoefficient[i].x-pBasisFunction[i].x*pFourierCoefficient[i].y;
		OutScatteringIntegral+=Summand*2.0f/Constant;
	}
	OutScatteringIntegral=saturate(1.0f-OutScatteringIntegral);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! This function evaluates the scattering integral resolving occlusions by means of 
   a rectified moment shadow map holding weighted prefix sums of rows.
  \param OutScatteringIntegral See ComputeSingleScatteringRadiance().
  \param RayEndTexCoord,RayDepth See 
		 ComputePrefilteredSingleScatteringCoordinatesLinearRectification().
  \param ScatteringIntegralWithoutOcclusion The output value in absence of 
		 occlusion.
  \param ShadowMapSampler A sampler for the rectified and prefiltered moment shadow 
		 map of the scene.
  \param MomentBias The bias used to overcome quantization artifacts in the moment 
		 shadow map.
  \param OverestimationWeight If this value is zero, the output underestimates the 
		 actual amount of single scattering. If it is one, it overestimates. 
		 Inbetween linear interpolation is employed.*/
void ComputePrefilteredSingleScattering4Moments(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D ShadowMapSampler0,float MomentBias,float OverestimationWeight=0.5f){
	float4 OptimizedSample=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	OutScatteringIntegral=1.0f-EstimateIntegralFrom4Moments(RayDepth,OptimizedSample,MomentBias,1.0f-OverestimationWeight);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! Like ComputePrefilteredSingleScattering4Moments() but uses six moments.*/
void ComputePrefilteredSingleScattering2_4Moments(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D_float2 ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,float MomentBias,float OverestimationWeight=0.5f){
	float2 OptimizedSample0=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	float4 OptimizedSample1=ShadowMapSampler1.Texture.Sample(ShadowMapSampler1.Sampler,RayEndTexCoord);
	OutScatteringIntegral=1.0f-EstimateIntegralFrom6Moments(RayDepth,OptimizedSample0,OptimizedSample1,MomentBias,1.0f-OverestimationWeight);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! Like ComputePrefilteredSingleScattering4Moments() but uses six moments.*/
void ComputePrefilteredSingleScattering3_3Moments(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D_float3 ShadowMapSampler0,SSamplerTexturePair2D_float3 ShadowMapSampler1,float MomentBias,float OverestimationWeight=0.5f){
	float3 OptimizedSample0=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	float3 OptimizedSample1=ShadowMapSampler1.Texture.Sample(ShadowMapSampler1.Sampler,RayEndTexCoord);
	OutScatteringIntegral=1.0f-EstimateIntegralFrom6Moments(RayDepth,OptimizedSample0.xy,float4(OptimizedSample0.z,OptimizedSample1),MomentBias,1.0f-OverestimationWeight);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! This function computes the number of thread groups to be dispatched for 
   generation of row-wise prefix sums.
  \param nOutThreadGroup The number of thread groups along each dimension to be 
		 dispatched for the horizontal filter.
  \param ThreadGroupSize The size of thread groups along each dimension for the 
		 prefix sum generation. Only the second entry is supposed to be not one.
  \param TextureSize The width and height of the texture for which prefix sums 
		 should be generated in x and y.*/
void GetPrefixSumThreadGroupCount(out uint3 nOutThreadGroup,uint3 ThreadGroupSize,float4 TextureSize){
	uint Height=(uint)(TextureSize.y);
	nOutThreadGroup=uint3(1,Height/ThreadGroupSize.y,1);
}


/*! Generates transmittance weighted horizontal prefix sums of the given source 
   texture and stores them in the given output texture (any four channel textures 
   are fine). Transmittance weighting means that summands are weighted by their 
   (approximate) contribution to the single scattering integral.
  \param OutFilteredShadowMap0 The texture which is to be overwritten by weighted 
		 prefix sums.
  \param ShadowMapSampler0 A texture and sampler providing access to the four 
		 channel shadow map for which prefix sums are to be generated. It must have 
		 the same dimensions as the output texture and it must be rectified in a 
		 linear fashion. Note that this method does not care about the contents of 
		 the texture.
  \param ThreadID The y component of this vector indicates the row on which to 
		 operate.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.
  \param ShadowMapProjectionToWorldSpace The projection to world space transform 
		 used for generation of the given shadow map.
  \param WorldSpaceCameraPosition The position of the camera which is to see the 
		 single scattering.*/
void ApplyTransmittanceWeightedPrefixSumLinearRectification(inout RWTexture2D<float4> OutFilteredShadowMap0,SSamplerTexturePair2D ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 ShadowMapProjectionToWorldSpace,float3 WorldSpaceCameraPosition){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Precompute quantities for the loop
	float ProjectionSpaceY=1.0f-2.0f*float(ThreadID.y)/float(Height);
	float InvHalfWidth=2.0f/float(Width);
	float4 WorldSpaceRayOffset=mul(float4(InvHalfWidth,0.0f,0.0f,0.0f),ShadowMapProjectionToWorldSpace);
	// Loop over texels in the row
	float PreviousIntegralValue=-1.0f;
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	float4 WorldSpaceRaySample=mul(float4(-0.5f*InvHalfWidth-1.0f,ProjectionSpaceY,0.5f,1.0f),ShadowMapProjectionToWorldSpace);
	for(uint x=0;x<Width;++x){
		// Compute a world space position that corresponds to the current texel
		WorldSpaceRaySample+=WorldSpaceRayOffset;
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*distance(WorldSpaceCameraPosition,WorldSpaceRaySample.xyz/WorldSpaceRaySample.w));
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		float4 ShadowMapSample=ShadowMapSampler0.Texture.Load(uint3(x,ThreadID.y,0));
		PrefixSum+=Weight*ShadowMapSample;
		TotalWeight+=Weight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum/TotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumLinearRectification() but for split 
   moment shadow maps with four channels total.*/
void ApplyTransmittanceWeightedPrefixSumLinearRectificationSplitMSM4(inout RWTexture2D<float4> OutFilteredShadowMap0,SSamplerTexturePair2D_float2 ShadowMapSampler0,SSamplerTexturePair2D_float2 ShadowMapSampler1,uint3 ThreadID,float ExtinctionCoefficient,float4x4 ShadowMapProjectionToWorldSpace,float3 WorldSpaceCameraPosition){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Precompute quantities for the loop
	float ProjectionSpaceY=1.0f-2.0f*float(ThreadID.y)/float(Height);
	float InvHalfWidth=2.0f/float(Width);
	float4 WorldSpaceRayOffset=mul(float4(InvHalfWidth,0.0f,0.0f,0.0f),ShadowMapProjectionToWorldSpace);
	// Loop over texels in the row
	float PreviousIntegralValue=-1.0f;
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	float4 WorldSpaceRaySample=mul(float4(-0.5f*InvHalfWidth-1.0f,ProjectionSpaceY,0.5f,1.0f),ShadowMapProjectionToWorldSpace);
	for(uint x=0;x<Width;++x){
		// Compute a world space position that corresponds to the current texel
		WorldSpaceRaySample+=WorldSpaceRayOffset;
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*distance(WorldSpaceCameraPosition,WorldSpaceRaySample.xyz/WorldSpaceRaySample.w));
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		float4 ShadowMapSample=float4(ShadowMapSampler0.Texture.Load(uint3(x,ThreadID.y,0)),ShadowMapSampler1.Texture.Load(uint3(x,ThreadID.y,0)));
		PrefixSum+=Weight*ShadowMapSample;
		TotalWeight+=Weight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum/TotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4() but 
   generates Fourier coefficients rather than moments optimized for quantization. 
   More precisely it generates Fourier coefficients of order FourierCoefficientOrder 
   and FourierCoefficientOrder+1 and stores them in rg and ba, respectively.*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationCSM(inout RWTexture2D<float4> OutFilteredShadowMap0,SSamplerTexturePair2D_float ShadowMapSampler0,float FourierCoefficientOrder,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	const float PI=3.1415926535897932384626433832795f;
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	float FirstFourierCoefficientFactor=FourierCoefficientOrder*2.0f*PI-PI;
	float SecondFourierCoefficientFactor=FirstFourierCoefficientFactor+2.0f*PI;
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Get the resampled depth value
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float ShadowMapDepth=ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f);
		float ResampledDepth=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		// Generate Fourier coefficients and prepare them for being stored as non-
		// negative values
		float4 FourierCoefficients;
		sincos(ResampledDepth*FirstFourierCoefficientFactor,FourierCoefficients.g,FourierCoefficients.r);
		sincos(ResampledDepth*SecondFourierCoefficientFactor,FourierCoefficients.a,FourierCoefficients.b);
		FourierCoefficients=FourierCoefficients*0.5f+0.5f;
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum+=Weight*FourierCoefficients;
		TotalWeight+=Weight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum/TotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Resamples the given common shadow map using the specified non-linear 
   rectification transform and generates four moments with optimized quantization on 
   the fly. These moments are stored as transmittance weighted prefix sums of the 
   rows of the image. See ApplyTransmittanceWeightedPrefixSumLinearRectification() 
   for details on transmittance weighting.
  \param OutFilteredShadowMap0 The texture which is to be overwritten by weighted 
		 prefix sums of moments.
  \param ShadowMapSampler0 A texture and sampler providing access to a common shadow 
		 map. The sampler should not use any filtering.
  \param ThreadID The y component of this vector indicates the row on which to 
		 operate.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.
  \param RectificationToShadowMapProjectionSpace A transformation from the linear 
		 space used to define the rectified coordinates to the projection space of 
		 the shadow map.
  \param DistanceMinMax,InclinationMinMax,AzimuthMinMax See 
		 GetRectifiedSpaceFrustumBounds().*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4(inout RWTexture2D<float4> OutFilteredShadowMap0,SSamplerTexturePair2D_float ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Get the resampled depth value
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float ShadowMapDepth=ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f);
		float ResampledDepth=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		// Generate moments that are optimized for quantization
		float4 Moments;
		ComputeMomentVector4MomentsOptimized(Moments,ResampledDepth);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum+=Weight*Moments;
		TotalWeight+=Weight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum/TotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4() but uses a 
   moment shadow map instead of a common shadow map as input. The moment shadow map 
   is sampled with filtering and moments are transformed in a sophisticated manner.*/
void ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationMSM4(inout RWTexture2D<float4> OutFilteredShadowMap0,SSamplerTexturePair2D ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Sample the moment shadow map
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float4 OptimizedMoments=ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f);
		// Represent the moments by a distribution consisting of two Dirac-deltas
		float2 ShadowMapDepth;
		float DepthWeight;
		GetSparseRepresentation4Moments(ShadowMapDepth,DepthWeight,OptimizedMoments);
		// Transform the depth values into the appropriate coordinate system and go 
		// back to a quantized vector of moments
		float2 ResampledDepth;
		ResampledDepth[0]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[0],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		ResampledDepth[1]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[1],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		float4 Moments0,Moments1;
		ComputeMomentVector4MomentsOptimized(Moments0,ResampledDepth[0]);
		ComputeMomentVector4MomentsOptimized(Moments1,ResampledDepth[1]);
		float4 Moments=lerp(Moments0,Moments1,DepthWeight);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum+=Weight*Moments;
		TotalWeight+=Weight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum/TotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationMSM4() but 
   uses a split moment shadow map as input.*/
void ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationSplitMSM4(inout RWTexture2D<float4> OutFilteredShadowMap0,SSamplerTexturePair2D_float2 ShadowMapSampler0,SSamplerTexturePair2D_float2 ShadowMapSampler1,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Sample the moment shadow map
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float4 OptimizedMoments=float4(
			ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f),
			ShadowMapSampler1.Texture.SampleLevel(ShadowMapSampler1.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f));
		// Represent the moments by a distribution consisting of two Dirac-deltas
		float2 ShadowMapDepth;
		float DepthWeight;
		GetSparseRepresentation4Moments(ShadowMapDepth,DepthWeight,OptimizedMoments);
		// Transform the depth values into the appropriate coordinate system and go 
		// back to a quantized vector of moments
		float2 ResampledDepth;
		ResampledDepth[0]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[0],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		ResampledDepth[1]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[1],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		float4 Moments0,Moments1;
		ComputeMomentVector4MomentsOptimized(Moments0,ResampledDepth[0]);
		ComputeMomentVector4MomentsOptimized(Moments1,ResampledDepth[1]);
		float4 Moments=lerp(Moments0,Moments1,DepthWeight);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum+=Weight*Moments;
		TotalWeight+=Weight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum/TotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4() but 
   produces six moments stored in a manner that is optimized for good quantization 
   in a two and a four channel texture (useful for 16- or 32-bit quantization).*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM2_4(inout RWTexture2D<float2> OutFilteredShadowMap0,inout RWTexture2D<float4> OutFilteredShadowMap1,SSamplerTexturePair2D_float ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float2 PrefixSum0=float2(0.0f,0.0f);
	float4 PrefixSum1=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Get the resampled depth value
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float ShadowMapDepth=ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f);
		float ResampledDepth=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		// Generate moments that are optimized for quantization
		float2 QuantizedMoments0;
		float4 QuantizedMoments1;
		ComputeMomentVector6MomentsOptimized(QuantizedMoments0,QuantizedMoments1,ResampledDepth);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum0+=Weight*QuantizedMoments0;
		PrefixSum1+=Weight*QuantizedMoments1;
		TotalWeight+=Weight;
		float InvTotalWeight=1.0f/TotalWeight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum0*InvTotalWeight;
		OutFilteredShadowMap1[uint2(x,ThreadID.y)]=PrefixSum1*InvTotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationMSM4() but 
   produces six moments stored in a manner that is optimized for good quantization 
   in a two and a four channel texture (useful for 16- or 32-bit quantization). The 
   input is a four moment shadow map nonetheless.*/
void ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationMSM2_4(inout RWTexture2D<float2> OutFilteredShadowMap0,inout RWTexture2D<float4> OutFilteredShadowMap1,SSamplerTexturePair2D ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float2 PrefixSum0=float2(0.0f,0.0f);
	float4 PrefixSum1=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Get the resampled depth value
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float4 OptimizedMoments=ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f);
		// Represent the moments by a distribution consisting of two Dirac-deltas
		float2 ShadowMapDepth;
		float DepthWeight;
		GetSparseRepresentation4Moments(ShadowMapDepth,DepthWeight,OptimizedMoments);
		// Transform the depth values into the appropriate coordinate system and go 
		// back to a quantized vector of moments
		float2 ResampledDepth;
		ResampledDepth[0]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[0],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		ResampledDepth[1]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[1],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		float2 Moments0_0,Moments1_0;
		float4 Moments0_1,Moments1_1;
		ComputeMomentVector6MomentsOptimized(Moments0_0,Moments0_1,ResampledDepth[0]);
		ComputeMomentVector6MomentsOptimized(Moments1_0,Moments1_1,ResampledDepth[1]);
		float2 QuantizedMoments0=lerp(Moments0_0,Moments1_0,DepthWeight);
		float4 QuantizedMoments1=lerp(Moments0_1,Moments1_1,DepthWeight);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum0+=Weight*QuantizedMoments0;
		PrefixSum1+=Weight*QuantizedMoments1;
		TotalWeight+=Weight;
		float InvTotalWeight=1.0f/TotalWeight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum0*InvTotalWeight;
		OutFilteredShadowMap1[uint2(x,ThreadID.y)]=PrefixSum1*InvTotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationMSM2_4() 
   but uses a split moment shadow map as input.*/
void ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationSplitMSM2_4(inout RWTexture2D<float2> OutFilteredShadowMap0,inout RWTexture2D<float4> OutFilteredShadowMap1,SSamplerTexturePair2D_float2 ShadowMapSampler0,SSamplerTexturePair2D_float2 ShadowMapSampler1,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float2 PrefixSum0=float2(0.0f,0.0f);
	float4 PrefixSum1=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Get the resampled depth value
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float4 OptimizedMoments=float4(
			ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f),
			ShadowMapSampler1.Texture.SampleLevel(ShadowMapSampler1.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f));
		// Represent the moments by a distribution consisting of two Dirac-deltas
		float2 ShadowMapDepth;
		float DepthWeight;
		GetSparseRepresentation4Moments(ShadowMapDepth,DepthWeight,OptimizedMoments);
		// Transform the depth values into the appropriate coordinate system and go 
		// back to a quantized vector of moments
		float2 ResampledDepth;
		ResampledDepth[0]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[0],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		ResampledDepth[1]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[1],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		float2 Moments0_0,Moments1_0;
		float4 Moments0_1,Moments1_1;
		ComputeMomentVector6MomentsOptimized(Moments0_0,Moments0_1,ResampledDepth[0]);
		ComputeMomentVector6MomentsOptimized(Moments1_0,Moments1_1,ResampledDepth[1]);
		float2 QuantizedMoments0=lerp(Moments0_0,Moments1_0,DepthWeight);
		float4 QuantizedMoments1=lerp(Moments0_1,Moments1_1,DepthWeight);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum0+=Weight*QuantizedMoments0;
		PrefixSum1+=Weight*QuantizedMoments1;
		TotalWeight+=Weight;
		float InvTotalWeight=1.0f/TotalWeight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum0*InvTotalWeight;
		OutFilteredShadowMap1[uint2(x,ThreadID.y)]=PrefixSum1*InvTotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM2_4() but 
   stores the six moments in two three-channel textures. This is useful for 10-bit 
   quantization.*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM3_3(inout RWTexture2D<float3> OutFilteredShadowMap0,inout RWTexture2D<float3> OutFilteredShadowMap1,SSamplerTexturePair2D_float ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float3 PrefixSum0=float3(0.0f,0.0f,0.0f);
	float3 PrefixSum1=float3(0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Get the resampled depth value
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float ShadowMapDepth=ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f);
		float ResampledDepth=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		// Generate moments that are optimized for quantization
		float2 QuantizedMoments0;
		float4 QuantizedMoments1;
		ComputeMomentVector6MomentsOptimized(QuantizedMoments0,QuantizedMoments1,ResampledDepth);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum0+=Weight*float3(QuantizedMoments0,QuantizedMoments1.x);
		PrefixSum1+=Weight*QuantizedMoments1.yzw;
		TotalWeight+=Weight;
		float InvTotalWeight=1.0f/TotalWeight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum0*InvTotalWeight;
		OutFilteredShadowMap1[uint2(x,ThreadID.y)]=PrefixSum1*InvTotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationMSM2_4() 
   but stores the six moments in two three-channel textures (useful for 10-bit 
   quantization).*/
void ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationMSM3_3(inout RWTexture2D<float3> OutFilteredShadowMap0,inout RWTexture2D<float3> OutFilteredShadowMap1,SSamplerTexturePair2D ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float3 PrefixSum0=float3(0.0f,0.0f,0.0f);
	float3 PrefixSum1=float3(0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Get the resampled depth value
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float4 OptimizedMoments=ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f);
		// Represent the moments by a distribution consisting of two Dirac-deltas
		float2 ShadowMapDepth;
		float DepthWeight;
		GetSparseRepresentation4Moments(ShadowMapDepth,DepthWeight,OptimizedMoments);
		// Transform the depth values into the appropriate coordinate system and go 
		// back to a quantized vector of moments
		float2 ResampledDepth;
		ResampledDepth[0]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[0],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		ResampledDepth[1]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[1],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		float2 Moments0_0,Moments1_0;
		float4 Moments0_1,Moments1_1;
		ComputeMomentVector6MomentsOptimized(Moments0_0,Moments0_1,ResampledDepth[0]);
		ComputeMomentVector6MomentsOptimized(Moments1_0,Moments1_1,ResampledDepth[1]);
		float2 QuantizedMoments0=lerp(Moments0_0,Moments1_0,DepthWeight);
		float4 QuantizedMoments1=lerp(Moments0_1,Moments1_1,DepthWeight);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum0+=Weight*float3(QuantizedMoments0,QuantizedMoments1.x);
		PrefixSum1+=Weight*QuantizedMoments1.yzw;
		TotalWeight+=Weight;
		float InvTotalWeight=1.0f/TotalWeight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum0*InvTotalWeight;
		OutFilteredShadowMap1[uint2(x,ThreadID.y)]=PrefixSum1*InvTotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationMSM3_3() 
   but uses a split moment shadow map as input.*/
void ApplyTransmittanceWeightedPrefixSumFilteredNonLinearRectificationSplitMSM3_3(inout RWTexture2D<float3> OutFilteredShadowMap0,inout RWTexture2D<float3> OutFilteredShadowMap1,SSamplerTexturePair2D_float2 ShadowMapSampler0,SSamplerTexturePair2D_float2 ShadowMapSampler1,uint3 ThreadID,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Prepare everything needed for the long loop in a manner that avoids 
	// computations within the loop as much as possible
	float3 TextureSpaceRaySample,TextureSpaceRayOffset;
	float2 ShadowMapToRectificationDepthTransform,InclinationToDepthTransform;
	float OffsetLength,TotalOffsetLength;
	PrepareShadowMapRowResampling(TextureSpaceRaySample,TextureSpaceRayOffset,ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,OffsetLength,TotalOffsetLength,uint2(Width,Height),ThreadID.y,RectificationToShadowMapProjectionSpace,DistanceMinMax,InclinationMinMax,AzimuthMinMax);
	// Loop over output texels in the row
	float PreviousIntegralValue=-1.0f;
	float3 PrefixSum0=float3(0.0f,0.0f,0.0f);
	float3 PrefixSum1=float3(0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	for(uint x=0;x<Width;++x){
		// Get the resampled depth value
		TextureSpaceRaySample+=TextureSpaceRayOffset;
		TotalOffsetLength+=OffsetLength;
		float4 OptimizedMoments=float4(
			ShadowMapSampler0.Texture.SampleLevel(ShadowMapSampler0.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f),
			ShadowMapSampler1.Texture.SampleLevel(ShadowMapSampler1.Sampler,TextureSpaceRaySample.xy/TextureSpaceRaySample.z,0.0f));
		// Represent the moments by a distribution consisting of two Dirac-deltas
		float2 ShadowMapDepth;
		float DepthWeight;
		GetSparseRepresentation4Moments(ShadowMapDepth,DepthWeight,OptimizedMoments);
		// Transform the depth values into the appropriate coordinate system and go 
		// back to a quantized vector of moments
		float2 ResampledDepth;
		ResampledDepth[0]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[0],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		ResampledDepth[1]=ConvertShadowMapDepthNonLinearRectification(ShadowMapDepth[1],ShadowMapToRectificationDepthTransform,InclinationToDepthTransform,TotalOffsetLength);
		float2 Moments0_0,Moments1_0;
		float4 Moments0_1,Moments1_1;
		ComputeMomentVector6MomentsOptimized(Moments0_0,Moments0_1,ResampledDepth[0]);
		ComputeMomentVector6MomentsOptimized(Moments1_0,Moments1_1,ResampledDepth[1]);
		float2 QuantizedMoments0=lerp(Moments0_0,Moments1_0,DepthWeight);
		float4 QuantizedMoments1=lerp(Moments0_1,Moments1_1,DepthWeight);
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*TotalOffsetLength);
		float Weight=IntegralValue-PreviousIntegralValue;
		// Contribute to the prefix sum
		PrefixSum0+=Weight*float3(QuantizedMoments0,QuantizedMoments1.x);
		PrefixSum1+=Weight*QuantizedMoments1.yzw;
		TotalWeight+=Weight;
		float InvTotalWeight=1.0f/TotalWeight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum0*InvTotalWeight;
		OutFilteredShadowMap1[uint2(x,ThreadID.y)]=PrefixSum1*InvTotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM2_4() but for 
   the case of a linear rectification transform.
  \sa ApplyTransmittanceWeightedPrefixSumLinearRectification() */
void ApplyTransmittanceWeightedPrefixSumLinearRectificationMSM2_4(inout RWTexture2D<float2> OutFilteredShadowMap0,inout RWTexture2D<float4> OutFilteredShadowMap1,SSamplerTexturePair2D_float ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 ShadowMapProjectionToWorldSpace,float3 WorldSpaceCameraPosition){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Precompute quantities for the loop
	float ProjectionSpaceY=1.0f-2.0f*float(ThreadID.y)/float(Height);
	float InvHalfWidth=2.0f/float(Width);
	float4 WorldSpaceRayOffset=mul(float4(InvHalfWidth,0.0f,0.0f,0.0f),ShadowMapProjectionToWorldSpace);
	// Loop over texels in the row
	float PreviousIntegralValue=-1.0f;
	float2 PrefixSum0=float2(0.0f,0.0f);
	float4 PrefixSum1=float4(0.0f,0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	float4 WorldSpaceRaySample=mul(float4(-0.5f*InvHalfWidth-1.0f,ProjectionSpaceY,0.5f,1.0f),ShadowMapProjectionToWorldSpace);
	for(uint x=0;x<Width;++x){
		// Compute a world space position that corresponds to the current texel
		WorldSpaceRaySample+=WorldSpaceRayOffset;
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*distance(WorldSpaceCameraPosition,WorldSpaceRaySample.xyz/WorldSpaceRaySample.w));
		float Weight=IntegralValue-PreviousIntegralValue;
		// Compute moments on the fly
		float ShadowMapDepth=ShadowMapSampler0.Texture.Load(uint3(x,ThreadID.y,0));
		float2 QuantizedMoments0;
		float4 QuantizedMoments1;
		ComputeMomentVector6MomentsOptimized(QuantizedMoments0,QuantizedMoments1,ShadowMapDepth);
		// Contribute to the prefix sum
		PrefixSum0+=Weight*QuantizedMoments0;
		PrefixSum1+=Weight*QuantizedMoments1;
		TotalWeight+=Weight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum0/TotalWeight;
		OutFilteredShadowMap1[uint2(x,ThreadID.y)]=PrefixSum1/TotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumLinearRectificationMSM2_4() but 
   stores the six moments in two three-channel textures. This is useful for 10-bit 
   quantization.*/
void ApplyTransmittanceWeightedPrefixSumLinearRectificationMSM3_3(inout RWTexture2D<float3> OutFilteredShadowMap0,inout RWTexture2D<float3> OutFilteredShadowMap1,SSamplerTexturePair2D_float ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float4x4 ShadowMapProjectionToWorldSpace,float3 WorldSpaceCameraPosition){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	// Precompute quantities for the loop
	float ProjectionSpaceY=1.0f-2.0f*float(ThreadID.y)/float(Height);
	float InvHalfWidth=2.0f/float(Width);
	float4 WorldSpaceRayOffset=mul(float4(InvHalfWidth,0.0f,0.0f,0.0f),ShadowMapProjectionToWorldSpace);
	// Loop over texels in the row
	float PreviousIntegralValue=-1.0f;
	float3 PrefixSum0=float3(0.0f,0.0f,0.0f);
	float3 PrefixSum1=float3(0.0f,0.0f,0.0f);
	float TotalWeight=0.0f;
	float4 WorldSpaceRaySample=mul(float4(-0.5f*InvHalfWidth-1.0f,ProjectionSpaceY,0.5f,1.0f),ShadowMapProjectionToWorldSpace);
	for(uint x=0;x<Width;++x){
		// Compute a world space position that corresponds to the current texel
		WorldSpaceRaySample+=WorldSpaceRayOffset;
		// Compute the contribution to the integral of this ray segment
		float IntegralValue=-exp(-ExtinctionCoefficient*distance(WorldSpaceCameraPosition,WorldSpaceRaySample.xyz/WorldSpaceRaySample.w));
		float Weight=IntegralValue-PreviousIntegralValue;
		// Compute moments on the fly
		float ShadowMapDepth=ShadowMapSampler0.Texture.Load(uint3(x,ThreadID.y,0));
		float2 QuantizedMoments0;
		float4 QuantizedMoments1;
		ComputeMomentVector6MomentsOptimized(QuantizedMoments0,QuantizedMoments1,ShadowMapDepth);
		// Contribute to the prefix sum
		PrefixSum0+=Weight*float3(QuantizedMoments0,QuantizedMoments1.x);
		PrefixSum1+=Weight*QuantizedMoments1.yzw;
		TotalWeight+=Weight;
		OutFilteredShadowMap0[uint2(x,ThreadID.y)]=PrefixSum0/TotalWeight;
		OutFilteredShadowMap1[uint2(x,ThreadID.y)]=PrefixSum1/TotalWeight;
		// Iterate on
		PreviousIntegralValue=IntegralValue;
	}
}
